@@ -9,6 +9,8 @@ class Expression(object):
     Expression objects are immutable.
     """
 
+    # {{{ arithmetic
+
     def __add__(self, other):
         if not is_valid_operand(other):
             return NotImplemented
@@ -91,6 +93,22 @@ class Expression(object):
         return quotient(other, self)
     __rtruediv__ = __rdiv__
 
+    def __floordiv__(self, other):
+        if not is_valid_operand(other):
+            return NotImplemented
+
+        if is_zero(other-1):
+            return self
+        return FloorDiv(self, other)
+
+    def __rfloordiv__(self, other):
+        if not is_valid_operand(other):
+            return NotImplemented
+
+        if is_zero(self-1):
+            return other
+        return FloorDiv(other, self)
+
     def __mod__(self, other):
         if not is_valid_operand(other):
             return NotImplemented
@@ -99,12 +117,12 @@ class Expression(object):
             return self
         return Remainder(self, other)
 
-    def __rmod(self, other):
+    def __rmod__(self, other):
         if not is_valid_operand(other):
             return NotImplemented
 
         return Remainder(other, self)
-    
+
     def __pow__(self, other):
         if not is_valid_operand(other):
             return NotImplemented
@@ -124,6 +142,8 @@ class Expression(object):
             return 1
         return Power(other, self)
 
+    # }}}
+
     def __neg__(self):
         return -1*self
 
@@ -131,8 +151,11 @@ class Expression(object):
         return Call(self, pars)
 
     def __getitem__(self, subscript):
-        return Subscript(self, subscript)
-    
+        if subscript == ():
+            return self
+        else:
+            return Subscript(self, subscript)
+
     def __float__(self):
         from pymbolic.mapper.evaluator import evaluate_to_float
         return evaluate_to_float(self)
@@ -150,7 +173,8 @@ class Expression(object):
 
         return "%s(%s)" % (self.__class__.__name__, initargs_str)
 
-    # hashable interface ------------------------------------------------------
+    # {{{ hashable interface
+
     def __eq__(self, other):
         """Provides equality testing with quick positive and negative paths
         based on L{id} and L{__hash__}().
@@ -180,19 +204,30 @@ class Expression(object):
             self.hash_value = self.get_hash()
             return self.hash_value
 
-    # hashable backend --------------------------------------------------------
+    # }}}
+
+    # {{{ hashable backend
+
     def is_equal(self, other):
-        raise NotImplementedError("is_equal() in "+str(type(self)))
+        return (type(other) == type(self)
+                and self.__getinitargs__() == other.__getinitargs__())
 
     def get_hash(self):
-        raise NotImplementedError("get_hash() in "+str(type(self)))
+        return hash((type(self),)+ self.__getinitargs__())
+
+    # }}}
+
+    # {{{ comparison interface
+
+    # /!\ Don't be tempted to resolve these to ComparisonOperator.
 
-    # comparison interface ----------------------------------------------------
     def __le__(self, other): raise TypeError("expressions don't have an order")
     def __lt__(self, other): raise TypeError("expressions don't have an order")
     def __ge__(self, other): raise TypeError("expressions don't have an order")
     def __gt__(self, other): raise TypeError("expressions don't have an order")
 
+    # }}}
+
 
 
 
@@ -228,19 +263,18 @@ class Variable(Leaf):
         else:
             return NotImplemented
 
-    def is_equal(self, other):
-        return (other.__class__ == self.__class__
-                and self.name == other.name)
-
-    def get_hash(self):
-        return hash((self.__class__, self.name))
-
-    def get_mapper_method(self, mapper):
-        return mapper.map_variable
+    mapper_method = intern("map_variable")
 
 
 
 
+class Wildcard(Leaf):
+    def __getinitargs__(self):
+        return ()
+
+    mapper_method = intern("map_wildcard")
+
+
 class FunctionSymbol(AlgebraicLeaf):
     """Represents the name of a function.
 
@@ -251,19 +285,14 @@ class FunctionSymbol(AlgebraicLeaf):
     def __getinitargs__(self):
         return ()
 
-    def is_equal(self, other):
-        return self.__class__ == other.__class
-
-    def get_hash(self):
-        return hash(self.__class__)
-
-    def get_mapper_method(self, mapper):
-        return mapper.map_function_symbol
+    mapper_method = intern("map_function_symbol")
 
 
 
 
 
+# {{{ structural primitives
+
 class Call(AlgebraicLeaf):
     def __init__(self, function, parameters):
         self.function = function
@@ -282,16 +311,7 @@ class Call(AlgebraicLeaf):
     def __getinitargs__(self):
         return self.function, self.parameters
 
-    def is_equal(self, other):
-        return isinstance(other, Call) \
-               and (self.function == other.function) \
-               and (self.parameters == other.parameters)
-
-    def get_hash(self):
-        return hash((self.__class__, self.function, self.parameters))
-
-    def get_mapper_method(self, mapper):
-        return mapper.map_call
+    mapper_method = intern("map_call")
 
 
 
@@ -304,17 +324,8 @@ class Subscript(AlgebraicLeaf):
     def __getinitargs__(self):
         return self.aggregate, self.index
 
-    def is_equal(self, other):
-        return isinstance(other, Subscript) \
-               and (self.aggregate == other.aggregate) \
-               and (self.index == other.index)
+    mapper_method = intern("map_subscript")
 
-    def get_hash(self):
-        return hash((self.__class__, self.aggregate, self.index))
-
-    def get_mapper_method(self, mapper):
-        return mapper.map_subscript
-        
 
 
 
@@ -326,19 +337,11 @@ class Lookup(AlgebraicLeaf):
     def __getinitargs__(self):
         return self.aggregate, self.name
 
-    def is_equal(self, other):
-        return isinstance(other, Lookup) \
-               and (self.aggregate == other.aggregate) \
-               and (self.name == other.name)
-
-    def get_hash(self):
-        return hash((self.__class__, self.aggregate, self.name))
-
-    def get_mapper_method(self, mapper):
-        return mapper.map_lookup
-
+    mapper_method = intern("map_lookup")
 
+# }}}
 
+# {{{ arithmetic primitives
 
 class Sum(Expression):
     def __init__(self, children):
@@ -349,10 +352,6 @@ class Sum(Expression):
     def __getinitargs__(self):
         return self.children
 
-    def is_equal(self, other):
-        return (isinstance(other, Sum) 
-                and (set(self.children) == set(other.children)))
-
     def __add__(self, other):
         if not is_valid_operand(other):
             return NotImplemented
@@ -390,11 +389,7 @@ class Sum(Expression):
             # FIXME: Right semantics?
             return True
 
-    def get_hash(self):
-        return hash((self.__class__, self.children))
-
-    def get_mapper_method(self, mapper):
-        return mapper.map_sum
+    mapper_method = intern("map_sum")
 
 
 
@@ -407,10 +402,6 @@ class Product(Expression):
     def __getinitargs__(self):
         return self.children
 
-    def is_equal(self, other):
-        return (isinstance(other, Product) 
-                and (set(self.children) == set(other.children)))
-
     def __mul__(self, other):
         if not is_valid_operand(other):
             return NotImplemented
@@ -439,11 +430,7 @@ class Product(Expression):
                 return False
         return True
 
-    def get_hash(self):
-        return hash((self.__class__, self.children))
-
-    def get_mapper_method(self, mapper):
-        return mapper.map_product
+    mapper_method = intern("map_product")
 
 
 
@@ -467,9 +454,6 @@ class QuotientBase(Expression):
     def __nonzero__(self):
         return bool(self.numerator)
 
-    def get_hash(self):
-        return hash((self.__class__, self.numerator, self.denominator))
-
 
 
 
@@ -480,21 +464,19 @@ class Quotient(QuotientBase):
                and (self.numerator == other.numerator) \
                and (self.denominator == other.denominator)
 
-    def get_mapper_method(self, mapper):
-        return mapper.map_quotient
+    mapper_method = intern("map_quotient")
+
+
+
+
+class FloorDiv(QuotientBase):
+    mapper_method = intern("map_floor_div")
 
 
 
 
 class Remainder(QuotientBase):
-    def is_equal(self, other):
-        from pymbolic.rational import Rational
-        return self.__class__ == other.__class__ \
-               and (self.numerator == other.numerator) \
-               and (self.denominator == other.denominator)
-
-    def get_mapper_method(self, mapper):
-        return mapper.map_remainder
+    mapper_method = intern("map_remainder")
 
 
 
@@ -507,20 +489,124 @@ class Power(Expression):
     def __getinitargs__(self):
         return self.base, self.exponent
 
-    def is_equal(self, other):
-        return isinstance(other, Power) \
-               and (self.base == other.base) \
-               and (self.exponent == other.exponent)
+    mapper_method = intern("map_power")
 
-    def get_hash(self):
-        return hash((self.__class__, self.base, self.exponent))
+# }}}
 
-    def get_mapper_method(self, mapper):
-        return mapper.map_power
+# {{{ comparisons, logic, conditionals
+
+class ComparisonOperator(Expression):
+    """Note: comparisons are not implicitly constructed by comparing
+    Expression objects.
+    """
+
+    def __init__(self, left, operator, right):
+        self.left = left
+        self.right = right
+        if not operator in [">", ">=", "==", "!=", "<", "<="]:
+            raise RuntimeError("invalid operator")
+        self.operator = operator
+
+    def __getinitargs__(self):
+        return self.left, self.operator, self.right
+
+    mapper_method = intern("map_comparison")
 
 
 
 
+class BooleanExpression(Expression):
+    pass
+
+class LogcialNot(BooleanExpression):
+    def __init__(self, child):
+        self.child = child
+
+    def __getinitargs__(self):
+        return (self.child, self.prefix)
+
+    mapper_method = intern("map_logical_not")
+
+
+
+
+class LogicalOr(BooleanExpression):
+    def __init__(self, children):
+        assert isinstance(children, tuple)
+
+        self.children = children
+
+    def __getinitargs__(self):
+        return self.children
+
+    mapper_method = intern("map_logical_or")
+
+
+
+
+class LogicalAnd(BooleanExpression):
+    def __init__(self, children):
+        assert isinstance(children, tuple)
+
+        self.children = children
+
+    def __getinitargs__(self):
+        return self.children
+
+    mapper_method = intern("map_logical_and")
+
+
+
+
+class If(Expression):
+    def __init__(self, criterion, then, else_):
+        self.condition = criterion
+        self.then = then
+        self.else_ = else_
+
+    def __getinitargs__(self):
+        return self.condition, self.then, self.else_
+
+    mapper_method = intern("map_if")
+
+
+
+
+class IfPositive(Expression):
+    def __init__(self, criterion, then, else_):
+        from warnings import warn
+        warn("IfPositive is deprecated, use If( ... >0)", DeprecationWarning,
+                stacklevel=2)
+
+        self.criterion = criterion
+        self.then = then
+        self.else_ = else_
+
+    def __getinitargs__(self):
+        return self.criterion, self.then, self.else_
+
+    mapper_method = intern("map_if_positive")
+
+
+
+
+class _MinMaxBase(Expression):
+    def __init__(self, children):
+        self.children = children
+
+    def __getinitargs__(self):
+        return self.children
+
+class Min(_MinMaxBase):
+    mapper_method = intern("map_min")
+
+class Max(_MinMaxBase):
+    mapper_method = intern("map_max")
+
+# }}}
+
+# {{{
+
 class Vector(Expression):
     """An immutable sequence that you can compute with."""
 
@@ -573,9 +659,11 @@ class Vector(Expression):
         return Vector(tuple(other*x for x in self))
 
     def __div__(self, other):
+        import operator
         return Vector(tuple(operator.div(x, other) for x in self))
 
     def __truediv__(self, other):
+        import operator
         return Vector(tuple(operator.truediv(x, other) for x in self))
 
     def __floordiv__(self, other):
@@ -584,11 +672,7 @@ class Vector(Expression):
     def __getinitargs__(self):
         return self.children
 
-    def get_hash(self):
-        return hash((self.__class__, self.children))
-
-    def get_mapper_method(self, mapper):
-        return mapper.map_vector
+    mapper_method = intern("map_vector")
 
 
 
@@ -601,52 +685,95 @@ class CommonSubexpression(Expression):
     def __getinitargs__(self):
         return (self.child, self.prefix)
 
-    def get_hash(self):
-        return hash((self.__class__, self.child))
-
-    def is_equal(self, other):
-        return (other.__class__ == self.__class__
-                and other.child == self.child)
-
     def get_extra_properties(self):
+        """Return a dictionary of extra kwargs to be passed to the
+        constructor from the identity mapper.
+
+        This allows derived classes to exist without having to
+        extend every mapper that processes them.
+        """
+
         return {}
 
-    def get_mapper_method(self, mapper): 
-        return mapper.map_common_subexpression
+    mapper_method = intern("map_common_subexpression")
 
 
 
+class Substitution(Expression):
+    """Work-alike of sympy's Subs."""
 
-
-class IfPositive(Expression):
-    def __init__(self, criterion, then, else_):
-        self.criterion = criterion
-        self.then = then
-        self.else_ = else_
+    def __init__(self, child, variables, values):
+        self.child = child
+        self.variables = variables
+        self.values = values
 
     def __getinitargs__(self):
-        return self.criterion, self.then, self.else_
+        return (self.child, self.variables, self.values)
 
-    def is_equal(self, other):
-        return (isinstance(other, IfPositive)
-                and self.criterion == other.criterion
-                and self.then == other.then
-                and self.else_ == other.else_)
-
-    def get_hash(self):
-        return hash((
-                self.__class__,
-                self.criterion,
-                self.then,
-                self.else_))
-
-    def get_mapper_method(self, mapper):
-        return mapper.map_if_positive
+    mapper_method = intern("map_substitution")
 
 
 
 
-# intelligent makers ---------------------------------------------------------
+class Derivative(Expression):
+    """Work-alike of sympy's Derivative."""
+
+    def __init__(self, child, variables):
+        self.child = child
+        self.variables = variables
+
+    def __getinitargs__(self):
+        return (self.child, self.variables)
+
+    mapper_method = intern("map_derivative")
+
+
+
+
+class Slice(Expression):
+    """A slice expression as in a[1:7]."""
+
+    def __init__(self, children):
+        assert isinstance(children, tuple)
+        self.children = children
+
+        if len(children) > 3:
+            raise ValueError("slice with more than three arguments")
+
+    def __getinitargs__(self):
+        return (self.children,)
+
+    def __nonzero__(self):
+        return True
+
+    @property
+    def start(self):
+        if len(self.children) > 1:
+            return self.children[0]
+        else:
+            return None
+
+    @property
+    def stop(self):
+        if len(self.children) == 1:
+            return self.children[0]
+        elif len(self.children) > 1:
+            return self.children[1]
+        else:
+            return None
+
+    @property
+    def step(self):
+        if len(self.children) == 3:
+            return self.children[2]
+        else:
+            return None
+
+    mapper_method = intern("map_slice")
+
+# }}}
+
+# intelligent factory functions ----------------------------------------------
 def make_variable(var_or_string):
     if not isinstance(var_or_string, Expression):
         return Variable(var_or_string)
@@ -663,15 +790,27 @@ def subscript(expression, index):
 
 
 def flattened_sum(components):
-    it = components.__iter__()
-    try:
-        result = it.next()
-    except StopIteration:
-        return 0
+    # flatten any potential sub-sums
+    queue = list(components)
+    done = []
 
-    for i in it:
-        result = result + i
-    return result
+    while queue:
+        item = queue.pop(0)
+
+        if is_zero(item):
+            continue
+
+        if isinstance(item, Sum):
+            queue += item.children
+        else:
+            done.append(item)
+
+    if len(done) == 0:
+        return 0
+    elif len(done) == 1:
+        return done[0]
+    else:
+        return Sum(tuple(done))
 
 
 
@@ -741,7 +880,7 @@ def quotient(numerator, denominator):
 
 
 
-# tool functions --------------------------------------------------------------
+# {{{ tool functions --------------------------------------------------------------
 global VALID_CONSTANT_CLASSES
 global VALID_OPERANDS
 VALID_CONSTANT_CLASSES = (int, float, complex)
@@ -791,3 +930,85 @@ def is_zero(value):
 
 
 
+
+def wrap_in_cse(expr, prefix=None):
+    if isinstance(expr, (Variable, Subscript)):
+        return expr
+
+    if isinstance(expr, CommonSubexpression):
+        if prefix is None:
+            return expr
+        if expr.prefix is None and type(expr) is CommonSubexpression:
+            return CommonSubexpression(expr.child, prefix)
+
+        # existing prefix wins
+        return expr
+
+    else:
+        return CommonSubexpression(expr, prefix)
+
+
+
+
+
+def make_common_subexpression(field, prefix=None):
+    try:
+        from pytools.obj_array import log_shape
+    except ImportError:
+        have_obj_array = False
+    else:
+        have_obj_array = True
+
+    if have_obj_array:
+        ls = log_shape(field)
+
+    if have_obj_array and ls != ():
+        from pytools import indices_in_shape
+        result = numpy.zeros(ls, dtype=object)
+
+        for i in indices_in_shape(ls):
+            if prefix is not None:
+                component_prefix = prefix+"_".join(str(i_i) for i_i in i)
+            else:
+                component_prefix = None
+
+            if is_constant(field[i]):
+                result[i] = field[i]
+            else:
+                result[i] = CommonSubexpression(field[i], component_prefix)
+
+        return result
+    else:
+        if is_constant(field):
+            return field
+        else:
+            return CommonSubexpression(field, prefix)
+
+
+
+
+def make_sym_vector(name, components):
+    """Return an object array of *components* subscripted
+    :class:`Field` instances.
+
+    :param components: The number of components in the vector.
+    """
+    if isinstance(components, int):
+        components = range(components)
+
+    from pytools.obj_array import join_fields
+    vfld = Variable(name)
+    return join_fields(*[vfld[i] for i in components])
+
+
+
+
+def variables(s):
+    return [Variable(s_i) for s_i in s.split() if s_i]
+
+# }}}
+
+
+
+
+# vim: foldmethod=marker
